class Py_cordEasyModifiedViews(Exception):
    """
    Main class exception
    """


class CustomIDNotFound(Py_cordEasyModifiedViews):
    def __init__(self):
        super().__init__(f"custom_id not found !")

class CoroutineError(Py_cordEasyModifiedViews):
    def __init__(self, coro):
        super().__init__(f"{coro} is not a coroutine !")



class MultibotError(Exception):
    pass

class BotAlreadyExistError(MultibotError):
    def __init__(self, bot_name: str):
        super().__init__(f"'{bot_name}' bot already exist !")

class BotNotFoundError(MultibotError):
    def __init__(self, bot_name: str):
        super().__init__(f"'{bot_name}' bot doesn't exist !")

class BotNotStartedError(MultibotError):
    def __init__(self, bot_name: str):
        super().__init__(f"'{bot_name}' not started !")

class BotNotReadyedError(MultibotError):
    def __init__(self, bot_name: str):
        super().__init__(f"'{bot_name}' not ready !")

class BotAlreadyStartedError(MultibotError):
    def __init__(self, bot_name: str):
        super().__init__(f"'{bot_name}' already started !")

class SetupCommandFunctionNotFound(MultibotError):
    def __init__(self, setup_command_name: str, file: str):
        super().__init__(f"'{setup_command_name}' function not found in '{file}' file ! Init commands impossible.")

class CommandFileNotFoundError(MultibotError):
    def __init__(self, file_name: str):
        super().__init__(f"'{file_name}' file not found ! Init commands impossible !")

class ModuleNotFoundError(MultibotError):
    def __init__(self, module_name: str):
        super().__init__(f"{module_name} module not found !")




from ..views.easy_modified_view import EasyModifiedViews
from .menu import Menu

from typing import Union, Callable, Any, Optional
from discord.components import ComponentType
from discord import ChannelType, Member, ApplicationContext, Interaction, Role
from discord.abc import GuildChannel

class SelectMenu:
    """
    Create a simply select menu
    """

    def __init__(self, timeout: Union[float, None] = None, disabled_on_timeout: bool = False):
        """
        Init the select menu
        """
        self.__select_menu: EasyModifiedViews = EasyModifiedViews(timeout=timeout, disabled_on_timeout=disabled_on_timeout)

    def add_string_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add a string select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.string_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_user_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add an user select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.user_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_role_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add a role select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.role_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_mentionable_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None) -> Menu:
        """
        Add a mentionable select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        """
        return self.__global_add_component(ComponentType.mentionable_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row)

    def add_channel_select_menu(self, custom_id: str = None, placeholder: str = None, min_values: int = 1, max_values: int = 1, disabled=False, row=None, channel_types: list[ChannelType] = None):
        """
        Add a channel select menu in the ui
        :param custom_id: The ID of the select menu that gets received during an interaction. If not given then one is generated for you.
        :param placeholder: The placeholder text that is shown if nothing is selected, if any.
        :param max_values: The maximum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param min_values: The minimum number of items that must be chosen for this select menu. Defaults to 1 and must be between 1 and 25.
        :param disabled: Whether the select is disabled or not.
        :param row: The relative row this select menu belongs to. A Discord component can only have 5 rows. By default, items are arranged automatically into those 5 rows. If you’d like to control the relative positioning of the row then passing an index is advised. For example, row=1 will show up before row=2. Defaults to None, which is automatic ordering. The row number must be between 0 and 4 (i.e. zero indexed).
        :param channel_types: A list of channel types that can be selected in this menu.
        """
        return self.__global_add_component(ComponentType.channel_select, custom_id=custom_id, placeholder=placeholder, max_values=max_values, min_values=min_values, disabled=disabled, row=row, channel_types=channel_types)

    def __global_add_component(self, component_type: ComponentType,
                               custom_id: Union[str, None] = None,
                               placeholder: Union[str, None] = None,
                               min_values: int = 1,
                               max_values: int = 1,
                               disabled: bool = False,
                               row: Union[int, None] = None,
                               channel_types: Union[ChannelType, None] = None) -> Menu:
        """
        global function to add a Select component
        """
        menu = Menu(component_type,
                    self,
                    **{'custom_id': custom_id, 'placeholder': placeholder, 'min_values': min_values,
                       'max_values': max_values, 'disabled': disabled, 'row': row, 'channel_types': channel_types})

        self.__select_menu.add_items(menu.component)

        return menu

    def set_callable(self, *custom_ids: str,
                     _callable: Union[Callable, None],
                     data: Optional[dict[str, Any]] = None,
                     autorised_roles : Optional[list[Union[int, Role]]] = None,
                     autorised_key: Optional[Callable] = None) -> "SelectMenu":
        """
        Set a callable for menus associated with custom_ids
        :param custom_ids: IDs menus
        :param _callable: The coroutine to set for all menus
        :param data: Add any data to pass in called function.
        :param autorised_roles: Any role ID allowed to interact with the view
        :param autorised_key: Callable function to check anything. The function get the current interaction passed in parameter
        """
        self.__select_menu.set_callable(*custom_ids, _callable=_callable, data=data, autorised_key=autorised_key, autorised_roles=autorised_roles)
        return self

    async def respond(self, ctx: Union[ApplicationContext, Interaction], *args, **kwargs) -> Any:
        """
        Respond at the ApplicationContext
        """
        return await self.__select_menu.respond(ctx=ctx, *args, view=self.__select_menu, **kwargs)

    async def send(self, target: Union[Member, GuildChannel], *args, **kwargs) -> Any:
        """
        Send at the target
        """
        return await self.__select_menu.send(target=target, *args, view=self.__select_menu, **kwargs)

    async def update(self):
        """
        Update the message.
        If the command doesn't have been respond, nothing happened.
        """
        if self.get_view.message is None:
            return
        await self.__select_menu.update_items(*self.get_view.items)

    def get_callable(self, custom_id: str) -> Union[Callable, None]:
        """
        Get the callable UI
        :param custom_id: UI ID
        """
        return self.__select_menu.get_callable(custom_id)

    @property
    def get_view(self) -> EasyModifiedViews:
        """
        Get the current view
        """
        return self.__select_menu




from ..views.easy_modified_view import EasyModifiedViews
from ..views.errors import CoroutineError
from discord.ui.button import Button, ButtonStyle
from discord import Emoji, PartialEmoji, Embed, Color
from typing import Optional, Callable, Union
from asyncio import iscoroutinefunction


class Poll:

    def __init__(self, timeout: Optional[float] = None, unique_vote: bool = True):
        """
        Class instance to make poll with buttons
        :param timeout: Time before end the view
        :param unique_vote: Set True to make vote unchangeable
        """
        self.yes_count: int = 0
        self.no_count: int = 0
        self.__unique_vote: bool = unique_vote
        self.__view = EasyModifiedViews(disabled_on_timeout=True, timeout=timeout)
        self.__button_yes = Button(label='Yes', emoji='✅', custom_id='Poll_yes', style=ButtonStyle.green)
        self.__button_no = Button(label='No', emoji='❌', custom_id='Poll_no', style=ButtonStyle.gray)
        self.__view.add_items(self.__button_yes, self.__button_no)
        self.__view.set_callable('Poll_yes', _callable=self.yes)
        self.__view.set_callable('Poll_no', _callable=self.no)
        self.__view.call_on_timeout(self._result) # asynchrone fonction quand la view est arrivé à son timeout
        self.__clicked_members: list[int] = []

    def add_answer_button(self, label: str,
                            custom_id: str,
                            emoji: Optional[Union[str, Emoji, PartialEmoji]] = None,
                            style: ButtonStyle = ButtonStyle.secondary,
                            row: Optional[int] = None,
                            _callable: Optional[Callable] = None) -> Button:
        """
        Create a new answer button and add it in the view.
        :param label: Button label
        :param custom_id: Button ID
        :param emoji: Button emoji
        :param style: Button style
        :param row: Button row
        :param _callable: Asynchronous function linked to the button interaction. It's called when the button is pressed
        """

        if not iscoroutinefunction(_callable):
            raise CoroutineError(_callable)

        b = Button(label=label, emoji=emoji, style=style, row=row, custom_id=custom_id)
        self.__view.add_items(b)
        self.__view.set_callable(custom_id, _callable=_callable)

        return b


    def set_yes_button(self, label: Optional[str] = None,
                       emoji: Optional[Union[str, Emoji, PartialEmoji]] = None,
                       style: Optional[ButtonStyle] = None,
                       row: Optional[int] = None,
                       _callable: Optional[Callable] = None):
        """
        Set yes button parameters (he didn't change dynamically if the view was sent before)
        :param label: Button label
        :param emoji: Button emoji
        :param style: Button style
        :param row: Button row
        :param _callable: Asynchronous function linked to the button interaction. It's called when the button is pressed
        """
        self.__button_yes.emoji = emoji if emoji is not None else self.__button_yes.emoji
        self.__button_yes.style = style if style is not None else self.__button_yes.style
        self.__button_yes.label = label if label is not None else self.__button_yes.label
        self.__button_yes.row = row if row is not None else self.__button_yes.row

        if not iscoroutinefunction(_callable):
            raise CoroutineError(_callable)

        if _callable is not None:
            self.__view.set_callable(self.__button_yes.custom_id, _callable=_callable)

    def set_no_button(self, label: Optional[str] = None,
                        emoji: Optional[Union[str, Emoji, PartialEmoji]] = None,
                        style: Optional[ButtonStyle] = None,
                        row: Optional[int] = None,
                        _callable: Optional[Callable] = None):
        """
        Set no button parameters (he didn't change dynamically if the view was sent before)
        :param label: Button label
        :param emoji: Button emoji
        :param row: Button row
        :param style: Button style
        :param _callable: Asynchronous function linked to the button interaction. It's called when the button is pressed
        """
        self.__button_no.emoji = emoji if emoji is not None else self.__button_no.emoji
        self.__button_no.style = style if style is not None else self.__button_no.style
        self.__button_no.label = label if label is not None else self.__button_no.label
        self.__button_no.row = row if row is not None else self.__button_no.row

        if not iscoroutinefunction(_callable):
            raise CoroutineError(_callable)

        if _callable is not None:
            self.__view.set_callable(self.__button_no.custom_id, _callable=_callable)

    async def yes(self, button, interaction, data):
        """
        Base asynchronous function when "yes" button is pressed.
        Increment “yes_count” attribute when pressed and respond with an ephemeral message.
        This function can be changed.
        """
        if self.__unique_vote and interaction.user.id in self.__clicked_members:
            await interaction.response.send_message(f"You have already voted !", ephemeral=True)
            return
        self.yes_count += 1
        self.__clicked_members.append(interaction.user.id)
        await interaction.response.send_message(f"You have selected : {button.label}", ephemeral=True)

    async def no(self, button, interaction, data):
        """
        Base asynchronous function when "no" button is pressed.
        Increment “yes_count” attribute when pressed and respond with an ephemeral message.
        This function can be changed.
        """
        if self.__unique_vote and interaction.user.id in self.__clicked_members:
            await interaction.response.send_message(f"You have already voted !", ephemeral=True)
            return
        self.yes_count += 1
        self.__clicked_members.append(interaction.user.id)
        await interaction.response.send_message(f"You have selected : {button.label}", ephemeral=True)

    async def _result(self, ctx):
        """
        Asynchronous base function called when timeout is reached.
        :param ctx: Actual command context
        """
        embed = Embed(title="Poll ended !",
                      description=f"## Results :\n"
                                  f" `-` {self.__button_yes.emoji.name} {self.__button_yes.label} -> `{self.yes_count}`\n"
                                  f" `-` {self.__button_no.emoji.name} {self.__button_no.label} -> `{self.no_count}`",
                      color=Color.gold())

        await self.__view.message.reply(embed=embed)

    @property
    def get_view(self) -> EasyModifiedViews:
        """
        Get the current view
        """
        return self.__view



from multiprocessing import Queue
from .errors import BotAlreadyExistError, BotNotFoundError, MultibotError, BotNotStartedError, ModuleNotFoundError
from .bot import DiscordBot
from discord import Intents
from immutableType import Str_
from sys import modules
from os import system
from psutil import Process

class ManageProcess:

    def __init__(self, main_queue: Queue, process_queue: Queue):
        """
        Gère tous les bots dans un processus
        """
        self.__bots: dict[str, DiscordBot] = {}
        self.main_queue: Queue = main_queue
        self.process_queue: Queue = process_queue
        self.removed_modules: dict[str, "ModuleType"] = {}
        self.allow_create_subprocess: bool = True
        self.process: Process = Process()

        self.commandes = {
            "ADD": self.add_bot_to_process,
            "REMOVE": self.remove_bot_to_process,
            "START": self.start_bot_to_process,
            "STOP": self.stop_bot_to_process,
            "RESTART": self.restart_bot_to_process,
            "IS_STARTED": self.is_started,
            "IS_READY": self.is_ready,
            "IS_WS_RATELIMITED": self.is_ws_ratelimited,
            "STOPALL": self.stop_all_bot_to_process,
            "STARTALL": self.start_all_bot_to_process,
            "RESTARTALL": self.restart_all_bot_to_process,
            "BOT_COUNT": self.bot_count,
            "STARTED_BOT_COUNT": self.started_bot_count,
            "SHUTDOWN_BOT_COUNT": self.shutdown_bot_count,
            "BOTS_bot_name": self.get_bots_bot_name,
            "RELOAD_COMMANDS": self.reload_all_commands,
            "ADD_COMMAND_FILE": self.add_pyFile_commands,
            "MODIFY_COMMAND_FILE": self.modify_pyFile_commands,
            "REMOVE_MODULES": self.remove_modules,
            "ADD_MODULES": self.add_modules,
            "ALLOW_SUBPROCESS": self.allow_subprocess
        }

    def run(self):
        """
        Boucle principale du processus, écoute la queue principale.
        Doit comporter aubligatoirement un dictionnaire avec la clé 'type'
        """
        wait_for = 0  # secondes avant de forcer la fermeture des subprocess
        while True:
            if not self.main_queue.empty():
                command: dict = self.main_queue.get()
                #print(command)

                type_request = command["type"]
                if type_request in self.commandes.keys():
                    del command['type']
                    try:
                        result = self.commandes[type_request](**command)
                        self.process_queue.put({'status': 'success', 'message': result, 'type': type_request})
                    except MultibotError as e:
                        self.process_queue.put({'status': 'error', 'message': e, 'type': type_request})

            if not self.allow_create_subprocess and wait_for > 50000 and (childrens_process := self.process.children()):
                for i in childrens_process:
                    try:
                        i.kill()
                    except Exception:
                        pass
                    finally:
                        wait_for = 0
            wait_for += 1

    def start_bot_to_process(self, bot_name: str) -> str:
        """
        Lance un unique bot
        """
        self.if_bot_no_exist(bot_name)
        self.__bots[bot_name].start()
        return f'{bot_name} bot started'

    def stop_bot_to_process(self, bot_name: str) -> str:
        """
        Stop un bot du processus
        :param bot_name: Le nom du bot à stopper
        """
        self.if_bot_no_exist(bot_name)
        self.__bots[bot_name].stop()
        return f'{bot_name} bot stopped'

    def restart_bot_to_process(self, bot_name: str) -> str:
        """
        Redémarre un bot du processus
        :param bot_name: Le nom du bot à redémarrer
        """
        self.if_bot_no_exist(bot_name)
        self.__bots[bot_name].stop()
        self.__bots[bot_name].start()
        self.__bots[bot_name].reload_pyFile_commands()
        return f'{bot_name} bot restarted'

    def restart_all_bot_to_process(self) -> list[str]:
        """
        Redémarre tous les bots du processus
        """
        result = []
        for bot in self.__bots.keys():
            result.append(self.restart_bot_to_process(bot))
        return result

    def start_all_bot_to_process(self) -> list[str]:
        """
        Start tous les bots du processus
        """
        result = []
        for bot in self.__bots.keys():
            result.append(self.start_bot_to_process(bot))
        return result

    def stop_all_bot_to_process(self) -> list[str]:
        """
        Stop tous les bots du processus
        """
        result = []
        for bot in self.__bots.keys():
            result.append(self.stop_bot_to_process(bot))

        return result

    def add_bot_to_process(self, bot_name: str, token: str, intents: Intents) -> str:
        """
        Ajoute un bot au processus
        :param bot_name: Le nom du bot
        :param token: Le token du bot
        :raise: BotAlreadyExistError si le bot existe déjà
        """
        if bot_name in self.__bots.keys():
            raise BotAlreadyExistError(bot_name)
        self.__bots[bot_name] = DiscordBot(token, intents)
        return f'Bot {bot_name} added'

    def add_pyFile_commands(self, bot_name: str, file: str, setup_function: str, reload_command: bool):
        """
        Ajoute et charge un fichier de commande bot et ses dépendances.
        Les fichiers doivent avoir une fonction appelée « setup » ou un équivalent passé en paramètre.

        def setup(bot: Bot) :
            ...

        :param bot_name : Le nom du bot à ajouter au fichier de commandes
        :param file: Chemin relatif ou absolue du fichier de commande
        :param setup_function : Nom de la fonction appelée par le processus pour donner l'instance de Bot.
        :param reload_command : Recharge toutes les commandes dans le fichier et les dépendances. Défaut : True
        """
        self.if_bot_no_exist(bot_name)
        file = Str_(file).str_
        setup_function = Str_(setup_function).str_
        self.__bots[bot_name].add_pyFile_commands(file=file, setup_function=setup_function, reload_command=reload_command)

    def modify_pyFile_commands(self, bot_name: str, file: str, setup_function: str):
        """
        Modifie un fichier de comandes et le recharge.
        Ne recharge que le fichier et non les commandes du bot !
        :param bot_name: Le nom du bot
        :param file: Le chemin d'accès relatif ou absolue du fichier
        """
        self.if_bot_no_exist(bot_name)
        file = Str_(file).str_
        self.__bots[bot_name].modify_pyFile_commands(file=file, setup_function=setup_function)

    def remove_pyFile_commands(self, bot_name: str, file: str):
        """
        Enlève un fichier de commandes et toutes ses commandes du bot.
        :param bot_name: Le nom du bot
        :param file: Le chemin d'accès relatif ou absolue du fichier
        """
        self.if_bot_no_exist(bot_name)
        file = Str_(file).str_
        self.__bots[bot_name].remove_pyFile_commands(file=file)

    def reload_all_commands(self, bot_name: str):
        """
        Recharge toutes les commandes sur Discord
        """
        self.if_bot_no_exist(bot_name)
        self.__bots[bot_name].reload_commands()
        return f'Bot {bot_name} commands reloaded'

    def remove_bot_to_process(self, bot_name: str) -> str:
        """
        Coupe et enlève un bot au processus
        :param bot_name: Le nom du bot à retirer
        :raise:
        """
        self.if_bot_no_exist(bot_name)
        try:
            self.__bots[bot_name].stop()
        except BotNotStartedError:
            pass
        self.__bots[bot_name].close_ascyncio_loop()
        del self.__bots[bot_name]
        return f'Bot {bot_name} removed'

    def remove_modules(self, modules_name: tuple[str]):
        """
        Enlève les modules (bibliothèque) de tous le processus (affecte donc les bots).
        A éxécuter avant de lancer un bot !
        :param modules_name: Tuple contenant les noms des modules à enlever
        """
        for module in modules_name:
            if module in modules_name:
                self.removed_modules[module] = modules.pop(module)
            else:
                system(f"pip uninstall {module}")
                #raise ModuleNotFoundError(module)
        return f"[{', '.join(modules_name)}] modules removed"

    def add_modules(self, modules_name: tuple[str]):
        """
        Ajoute les modules (bibliothèque) de tous le processus (affecte donc les bots).
        Uniquement les modules enlever au préalable peuvent-être de nouveau ajouter !
        A éxécuter avant de lancer un bot !
        :param modules_name: Tuple contenant les noms des modules à ajouter
        """
        for module in modules_name:
            if module in self.removed_modules.keys():
                modules[module] = self.removed_modules.pop(module)
            else:
                system(f"pip install {module}")
                #raise ModuleNotFoundError(module)
        return f"[{', '.join(modules_name)}] modules added"

    def is_started(self, bot_name: str) -> bool:
        """
        Regarde si la connexion au Websocket est effectué
        :param bot_name: Le nom du bot à vérifier
        """
        self.if_bot_no_exist(bot_name)
        return self.__bots[bot_name].is_running

    def is_ready(self, bot_name: str) -> bool:
        """
        Regarde si le bot est ready
        :param bot_name: Le nom du bot à vérifier
        """
        self.if_bot_no_exist(bot_name)
        return self.__bots[bot_name].is_ready

    def is_ws_ratelimited(self, bot_name: str) -> bool:
        """
        Regarde si le bot est ratelimit
        :param bot_name: Le nom du bot à vérifier
        """
        self.if_bot_no_exist(bot_name)
        return self.__bots[bot_name].is_ws_ratelimited

    def if_bot_no_exist(self, bot_name: str) -> None:
        """
        Regarde si le bot existe dans la class
        """
        if bot_name not in self.__bots.keys():
            raise BotNotFoundError(bot_name)

    def bot_count(self) -> int:
        """
        Renvoie le nombre de bot au total
        """
        return len(self.__bots)

    def started_bot_count(self) -> int:
        """
        Renvoie le nombre de bot démarré au total
        """
        s = 0
        for bot in self.__bots.values():
            if bot.is_running:
                s += 1
        return s

    def shutdown_bot_count(self) -> int:
        """
        Renvoie le nombre de bot arrêter au total
        """
        s = 0
        for bot in self.__bots.values():
            if not bot.is_running:
                s += 1
        return s

    def get_bots_bot_name(self) -> list[str]:
        """
        Renvoie tous les noms des bots entrée par l'utilisateur
        """
        return list(self.__bots.keys())

    def allow_subprocess(self, allow: bool) -> str:
        """
        Permet ou non l'utilisation de subprocess dans les bots.
        ATTENTION : Peut poser des problèmes de sécurité si un utilisateur malveillant à accès au code.
        Par défaut, False.
        :param allow: True pour autoriser, False pour interdire
        """
        self.allow_create_subprocess = allow
        return f'Subprocess allowed: {allow}'



from __future__ import annotations
from discord import Interaction, ApplicationContext, Message, Member, Role
from discord.abc import GuildChannel
from discord.ui import View, Item
from typing import Union, Callable, TYPE_CHECKING, Optional, Any
from asyncio import iscoroutinefunction, create_task

from .errors import CustomIDNotFound, CoroutineError

if TYPE_CHECKING:
    from ..menu.selectMenu import SelectMenu
    from ..pagination.pagination_view import Pagination
    from ..kit import Poll, Confirm


class EasyModifiedViews(View):
    """
    Class
    -------------
    Allows you to easily modify and replace an ui.
    """

    def __init__(self, timeout: Optional[float] = None, disabled_on_timeout: bool = False, call_on_timeout: Optional[Callable] = None, *items: Item):
        """
        Init a Class view for Discord UI
        :param timeout: The time before ui disable
        :param disabled_on_timeout: disable view if timeout is reached
        :param call_on_timeout: asynchronous function to call after view timed-out
        """
        super().__init__(*items, timeout=timeout)
        self.__timeout: Optional[float] = timeout
        self.__disabled_on_timeout: bool = disabled_on_timeout
        self.__callback: dict[str, dict[str, Union[Callable[[Interaction], None], Item, Any]]] = {}
        self.__ctx: Union[Message, Interaction] = None
        self.__call_on_timeout: Callable = call_on_timeout

    def __check_custom_id(self, custom_id: str) -> None:
        """
        Check if the custom_id is alive
        :param custom_id: ID to find
        :raise: CustomIDNotFound
        """
        if custom_id not in self.__callback.keys():
            raise CustomIDNotFound()

    async def respond(self, ctx: Union[ApplicationContext, Interaction], *args, **kwargs) -> None:
        """
        Respond at the ApplicationContext or Interaction
        """
        self.__ctx = await ctx.respond(*args, **kwargs)

    async def send(self, target: Union[Member, GuildChannel], *args, **kwargs) -> None:
        """
        Send at the target
        """
        self.__ctx = await target.send(*args, **kwargs)

    def add_items(self,
                   *items: Union[EasyModifiedViews, Confirm, Pagination, Poll, SelectMenu, Item]) -> EasyModifiedViews:
        """
        Add all items in the View.
        """

        for ui in items:

            if ui is None: # si l'ui n'est pas setup
                continue

            if type(ui).__name__ in ('SelectMenu', 'Pagination', 'Confirm', 'Poll', 'EasyModifiedViews'):
                for item in ui.get_view.items:
                    self.add_items(item)
                    self.set_callable(item.custom_id, _callable=ui.get_view.get_callable(item.custom_id), data=ui.get_view.get_callable_data(item.custom_id))

            else:
                self.__callback[ui.custom_id] = {'ui': ui, 'func': None, 'data': {}, 'autorised_roles': None, 'autorised_key': None}
                self.add_item(ui)

        return self

    async def update_items(self, *items: Item) -> "EasyModifiedViews":
        """
        Update all views.
        Append items if custom_ids not in the view
        Update items if custom_ids in the view
        :param items: items to update
        """

        for item in items:
            try:
                self.__check_custom_id(item.custom_id)
                self.__callback[item.custom_id]['ui'] = item

            except CustomIDNotFound:
                self.add_items(item)

        await self._update()
        return self

    def set_callable_decorator(self, custom_id: str,
                               data: Optional[dict[str, Any]] = None,
                               autorised_roles : Optional[list[Union[int, Role]]] = None,
                               autorised_key: Optional[Callable] = None):
        """
        Decorator to set up a callable for the item

        **Ui, Interaction and data parameters is required in coroutine function !**

        view = EasyModifiedViews(None)
        view.add_view(discord.ui.Button(label='coucou', custom_id='test_ID'))

        @view.set_callable_decorator(custom_id='test_ID')

        async def rep(**UI**, **interaction**, data):
            await interaction.response.send_message('coucou !!!')

        await ctx.respond('coucou', view=view)

        :param custom_id: item ID of the view
        :param data: Add any data to pass in called function.
        :param autorised_roles: Any role ID allowed to interact with the view
        :param autorised_key: Callable function to check anything. The function get the current interaction and data passed in parameter
        """

        def decorator(_callable: Callable):
            self.set_callable(custom_id,
                              _callable=_callable,
                              data=data,
                              autorised_roles=autorised_roles,
                              autorised_key=autorised_key)
            return _callable

        return decorator

    def set_callable(self, *custom_ids: str,
                     _callable: Callable,
                     data: Optional[dict[str, Any]] = None,
                     autorised_roles : Optional[list[Union[int, Role]]] = None,
                     autorised_key: Optional[Callable] = None):
        """
        set up a callable for items
        :param custom_ids: items IDs of the view
        :param _callable: The asynchronous callable linked. Take UI (Button, Select...) and Interaction parameters.
        :param data: Add any data to pass in called function.
        :param autorised_roles: Any role ID allowed to interact with the view
        :param autorised_key: Callable function to check anything. The function get the current interaction and data passed in parameter

        **UI, Interaction and data parameter is required in callable function !**

        view = EasyModifiedViews(None)

        view.add_view(discord.ui.Button(label='coucou', custom_id='test_ID'))

        async def rep(**UI**, **interaction**, data: dict[str, Any]):
            await interaction.response.send_message(data['message'])

        view.set_callable(custom_id='test_ID', callable=rep, data={'message': 'Hello !'})
        await ctx.respond('coucou', view=view)
        """
        for custom_id in custom_ids:
            self.__check_custom_id(custom_id)

            self.__callback[custom_id]['func'] = _callable
            self.__callback[custom_id]['data'] = data if data is not None else {}
            self.__callback[custom_id]['autorised_key'] = autorised_key
            if autorised_roles is not None:
                self.__callback[custom_id]['autorised_roles'] = [role.id if isinstance(role, Role) else role for role in autorised_roles if isinstance(role, (Role, int))]
            else:
                self.__callback[custom_id]['autorised_roles'] = None

    def get_callable(self, custom_id: str) -> Union[Callable, None]:
        """
        Get the callable UI
        :param custom_id: UI ID
        """
        self.__check_custom_id(custom_id)
        return self.__callback[custom_id]['func']

    async def interaction_check(self, interaction: Interaction) -> bool:
        """
        Func to apply items
        """
        func = self.__callback[interaction.custom_id]['func']
        data = self.__callback[interaction.custom_id]['data']
        autorised_roles = self.__callback[interaction.custom_id]['autorised_roles']
        autorised_key = self.__callback[interaction.custom_id]['autorised_key']

        if func is not None:
            if autorised_roles is not None: # si des rôles sont obligatoires
                if not (set(autorised_roles) & set([i.id for i in interaction.user.roles])): # si aucun rôle de l'user est dans les rôles aubligatoires
                    await interaction.response.defer(invisible=True)
                    return False

            if autorised_key is not None: # si une fonction est config pour la vérification
                if not autorised_key(interaction, data): # Si la fonction renvoie False
                    return False

            return await func(self.__callback[interaction.custom_id]['ui'], interaction, data)

        else:
            await interaction.response.defer(invisible=True)

        return True

    async def shutdown(self) -> None:
        """
        Disable all items (ui) in the view
        """
        self.disable_all_items()
        await self._update()

    async def disable_items(self, *custom_ids: str) -> None:
        """
        Disable partial items in the view
        :param custom_ids: custom ids of all items to deactivate
        """

        self.disable_all_items(exclusions=[self.get_item(id_) for id_ in self.__callback.keys() if id_ not in custom_ids])
        await self._update()

    async def enable_items(self, *custom_ids: str) -> None:
        """
        Enabl partial items in the view
        :param custom_ids: custom ids of all items to activate
        """
        self.enable_all_items(
            exclusions=[self.get_item(id_) for id_ in self.__callback.keys() if id_ not in custom_ids])
        await self._update()

    async def full_enable_items(self) -> None:
        """
        Enable all items in the view
        """
        self.enable_all_items()
        await self._update()


    async def switch_status_items(self):
        """
        Switch status for all items
        Enable -> Disable
        Disable -> Enable
        """

        for key, in self.__callback.keys():
            self.__callback[key]['ui'].disabled = not self.__callback[key]['ui'].disabled

        await self._update()


    def is_items_disabled(self, *custom_ids: str) -> bool:
        """
        Return True if all items are disabled
        """

        for custom_id in custom_ids:
            self.__check_custom_id(custom_id)
            if not self.__callback[custom_id]['ui'].disabled:
                return False

        return True


    def is_items_enabled(self, *custom_ids: str) -> bool:
        """
        Return True il aff items are enabled
        """

        for custom_id in custom_ids:
            self.__check_custom_id(custom_id)
            if self.__callback[custom_id]['ui'].disabled:
                return False

        return True

    async def delete_items(self, clear_all: bool = False, *custom_ids: str) -> None:
        """
        Delete an item on the view
        :param custom_ids: IDs of items to delete
        :param clear_all: Clear all items in the view
        :raise: CustomIDNotFound
        """
        if clear_all:
            self.clear_items()

        else:

            for custom_id in custom_ids:
                self.__check_custom_id(custom_id)
                self.remove_item(self.get_item(custom_id))

        await self._update()

    async def on_timeout(self) -> None:
        """
        Called if timeout view is finished
        """
        if self.__disabled_on_timeout:
            await self.shutdown()
        if self.__call_on_timeout is not None:
            await self.__call_on_timeout(self.__ctx)

    def call_on_timeout(self, _callable: Callable) -> None:
        """
        Call asynchronous function passed when the timeout is reached.
        :param _callable: asynchronous function to call. It takes one argument : ctx
        """
        if iscoroutinefunction(_callable):
            self.__call_on_timeout = _callable
        else:
            raise CoroutineError(_callable)

    async def _update(self) -> None:
        """
        Update the View on the attached message.
        """
        if self.is_finished() and not self.__disabled_on_timeout:
            return

        if self.message:
            await self.message.edit(view=self)

        elif self.__ctx:
            await self.__ctx.edit(view=self)

    def get_ui(self, custom_id: str) -> Item:
        """
        Get an ui in the view
        :param custom_id: UI ID
        :raise: CustomIDNotFound
        """
        self.__check_custom_id(custom_id)
        return self.__callback[custom_id]['ui']

    def get_callable_data(self, custom_id: str) -> dict[str, Any]:
        """
        Get datas passed in the callable asynchronous function
        :param custom_id: UI ID
        """
        self.__check_custom_id(custom_id)
        return self.__callback[custom_id]['data']

    def copy(self) -> EasyModifiedViews:
        """
        Return an exact copy of the view. All mutable object in the view is a new object
        """
        e = EasyModifiedViews(timeout=self.__timeout, disabled_on_timeout=self.__disabled_on_timeout).add_items(*self.items)
        for i in self.items:
            e.set_callable(i.custom_id, _callable=self.get_callable(i.custom_id), data=self.get_callable_data(i.custom_id))

        return e

    def __str__(self):
        return str(self.__callback)

    @property
    def items(self) -> tuple[Item]:
        return tuple([i['ui'] for i in self.__callback.values()])

    @property
    def get_view(self) -> EasyModifiedViews:
        """
        Get The current view. Don't use in your code
        """
        return self

    def __add__(self, _view: Union[EasyModifiedViews, Confirm, Pagination, Poll, SelectMenu, Item]) -> EasyModifiedViews:
        """
        Add all items to _view from the current EasyModifiedViews instance
        """
        self.add_items(_view)
        return self

    def __iadd__(self, _view) -> EasyModifiedViews:
        """
        Add all items to _view from the current EasyModifiedViews instance
        """
        return self.__add__(_view)



from setuptools import setup

with open("README.md", encoding='utf-8') as file:
    file = file.read()

setup(
    name="PycordViews",
    version="1.3.7.3",
    url="https://github.com/BOXERRMD/Py-cord_Views",
    author="Chronos (alias BOXERRMD)",
    author_email="vagabonwalybi@gmail.com",
    maintainer="Chronos",
    license="MIT License",
    description="Views and multibot for py-cord library",
    long_description=file,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows :: Windows 11",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.9"
    ],
    install_requires=[
        "immutable-Python-type",
        "psutil"
    ],
    packages=['pycordViews', 'pycordViews/pagination', 'pycordViews/views', 'pycordViews/menu', 'pycordViews/multibot', 'pycordViews/kit', 'pycordViews/modal'],
    python_requires='>=3.9'
)


from queue import Empty
from multiprocessing import get_context
from multiprocessing.queues import Queue
from .process import ManageProcess
from discord import Intents
from sys import platform
from typing import Union, Optional


class Multibot:

    def __init__(self, global_timeout: int = 30):
        """
        Get instance to run few Discord bot
        """
        if platform == 'win32':
            ctx = get_context("spawn")
        else:
            ctx = get_context("forkserver")
        self.__main_queue: Queue = ctx.Queue()
        self.__process_queue: Queue = ctx.Queue()
        # Création du processus gérant les bots
        self.__DiscordProcess = ctx.Process(target=self._start_process)
        self.__DiscordProcess.start()
        
        self.global_timeout = global_timeout
        
    def __get_data_queue(self, type: str) -> Union[list[dict], dict]:
        """
        Récupère les données dans la queue processus
        :param type: Type de la requête. Si elle ne correspond pas à la requête écoutée, elle refera un cicle d'écoute
        """
        try:
            result = self.__process_queue.get(timeout=self.global_timeout)
            if result['type'] == type:
                return result
            else:
                # Si le type ne correspond pas, on relance la récupération
                return self.__get_data_queue(type)
        except Empty:
            return {'status': 'error', 'message': 'timeout request exceeded'}
        except ValueError:
            return {'status': 'critical error', 'message': 'queue was closed ! Process was killed ?'}

    def _start_process(self):
        """
        Initialise et exécute le gestionnaire de processus.
        """
        manager = ManageProcess(self.__main_queue, self.__process_queue)
        manager.run()

    def add_bot(self, bot_name: str, token: str, intents: Intents):
        """
        Add a bot in the process
        :param bot_name: Bot name
        :param token: Token bot
        :param intents: Intents bot to Intents discord class
        """
        request_type = "ADD"
        self.__main_queue.put({"type": request_type,
                               "bot_name": bot_name,
                               "token": token,
                               "intents": intents})
        response = self.__get_data_queue(request_type)
        return response  # Retourne le statut de l'ajout

    def remove_bots(self, *bot_names: str) -> list[dict[str, str]]:
        """
        Shutdown and remove bots
        :param bot_names: Bot name to remove
        """
        request_type = "REMOVE"
        response = []
        for bot_name in bot_names:
            self.__main_queue.put({"type": request_type, "bot_name": bot_name})
            response.append(self.__get_data_queue(request_type))
        return response  # Retourne le statut de la suppression

    def start(self, *bot_names: str) -> list[dict[str, str]]:
        """
        Start bots
        :param bot_names: Bots name to start
        :return: List of data bot status
        """
        request_type = "START"
        results = []
        for bot_name in bot_names:
            self.__main_queue.put({'type': request_type, 'bot_name': bot_name})
            results.append(self.__get_data_queue(request_type))
        return results

    def stop(self, *bot_names: str) -> list[dict[str, str]]:
        """
        Stop bots
        :param bot_names: Bots name to start
        :return: Data status dict
        """
        request_type = "STOP"
        results = []
        for bot_name in bot_names:
            self.__main_queue.put({'type': request_type, 'bot_name': bot_name})
            results.append(self.__get_data_queue(request_type))
        return results

    def restart(self, *bot_names: str) -> list[dict[str, str]]:
        """
        Stop and start bots.
        This function is slow ! It's shutdown all bots properly.
        """
        request_type = "RESTART"
        results = []
        for bot_name in bot_names:
            self.__main_queue.put({'type': request_type, 'bot_name': bot_name})
            results.append(self.__get_data_queue(request_type))
        return results

    def restart_all(self):
        """
        Stop and restart all bots
        This function is slow ! It's shutdown all bots properly.
        """
        request_type = "RESTARTALL"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)

    def start_all(self) -> list[dict[str, list[str]]]:
        """
        Start all bots in the process.
        """
        request_type = "STARTALL"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)
    
    def stop_all(self) -> list[dict[str, list[str]]]:
        """
        Stop all bots in the process.
        This function is slow ! It's shutdown all bots properly.
        """
        request_type = "STOPALL"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)

    def add_modules(self, *modules_name):
        """
        Adds modules (library) to the process (thus affecting bots).
        Only previously removed modules can be added again!
        To be run before launching a bot!
        :param modules_name: names of modules to be added
        """
        request_type = "ADD_MODULES"
        self.__main_queue.put({'type': request_type, 'modules_name': modules_name})
        return self.__get_data_queue(request_type)

    def remove_modules(self, *modules_name):
        """
        Removes modules (library) to the process (thus affecting bots).
        To be run before launching a bot!
        :param modules_name: names of modules to be removed
        """
        request_type = "REMOVE_MODULES"
        self.__main_queue.put({'type': request_type, 'modules_name': modules_name})
        return self.__get_data_queue(request_type)

    def is_started(self, bot_name: str) -> bool:
        """
        Return the current Websocket connexion status
        :param bot_name: Bot name
        :return: True if the Websocket is online, else False
        """
        request_type = "IS_STARTED"
        self.__main_queue.put({'type': request_type, 'bot_name': bot_name})
        return self.__get_data_queue(request_type)['message']

    def is_ready(self, bot_name: str) -> bool:
        """
        Return the current bot connexion status
        :param bot_name: Bot name
        :return: True if the bot if ready, else False
        """
        request_type = "IS_READY"
        self.__main_queue.put({'type': request_type, 'bot_name': bot_name})
        return self.__get_data_queue(request_type)['message']

    def is_ws_ratelimited(self, bot_name: str) -> bool:
        """
        Get the current ratelimit status of the bot
        :param bot_name: Bot name
        :return: True if the bot was ratelimited, else False
        """
        request_type = "IS_WS_RATELIMITED"
        self.__main_queue.put({'type': request_type, 'bot_name': bot_name})
        return self.__get_data_queue(request_type)['message']

    def reload_commands(self, *bot_names: str) -> list[dict[str, str]]:
        """
        Reload all commands for each bot when bots are ready
        :param bot_names: Bots name to reload commands
        """
        request_type = "RELOAD_COMMANDS"
        result = []
        for name in bot_names:
            self.__main_queue.put({'type': request_type, 'bot_name': name})
            result.append(self.__get_data_queue(request_type))
        return result

    def add_pyFile_commands(self, bot_name: str, file: str, setup_function: str = 'setup', reload_command: bool = True) -> dict[str, str]:
        """
        Add and load a command bot file and dependencies.
        Files must have a function called ‘setup’ or an equivalent passed as a parameter.

        def setup(bot: Bot):
            ...

        :param bot_name: The bot's name to add commands file
        :param file: Relative or absolute commands file's path
        :param setup_function: Function name called by the process to give the Bot instance. Set to 'setup' by default.
        :param reload_command: Reload all command in the fil and dependencies. Default : True
        """
        request_type = "ADD_COMMAND_FILE"
        self.__main_queue.put({'type': request_type,
                               'bot_name': bot_name,
                               'file': file,
                               'setup_function': setup_function,
                               'reload_command': reload_command})
        return self.__get_data_queue(request_type)

    def modify_pyFile_commands(self, bot_name: str, file: str, setup_function: str = 'setup') -> dict[str, str]:

        """
        Modifies a file of commands and reloads it.
        Reloads only the file, not the bot commands!
        :param bot_name: The bot's name
        :param file: The file's relative or absolute path
        :param setup_function: Function name called by the process to give the Bot instance. Set to 'setup' by default.
        """
        request_type = "MODIFY_COMMAND_FILE"
        self.__main_queue.put({'type': request_type,
                               'bot_name': bot_name,
                               'file': file,
                               'setup_function': setup_function})
        return self.__get_data_queue(request_type)

    @property
    def bot_count(self) -> int:
        """
        Return the total number of bots
        """
        request_type = "BOT_COUNT"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)['message']

    @property
    def started_bot_count(self) -> int:
        """
        Return the total number of started bots
        """
        request_type = "STARTED_BOT_COUNT"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)['message']

    @property
    def shutdown_bot_count(self) -> int:
        """
        Return the total number of shutdown bots
        """
        request_type = "SHUTDOWN_BOT_COUNT"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)['message']

    @property
    def get_bots_name(self) -> list[str]:
        """
        Return all bots name (not real name of bots)
        """
        request_type = "BOTS_NAME"
        self.__main_queue.put({'type': request_type})
        return self.__get_data_queue(request_type)['message']
